"""Module for managing commands"""
import difflib
import inspect
import pickle
from dataclasses import dataclass
from functools import partial
from types import CodeType, NoneType
from typing import Callable, Iterable

from . import database
from jsonxx.listx import ListX


class AfterFunc(database.Struct):
    """Structure for AfterFuncs"""
    save_by = "user_id"
    user_id = ""
    after_name = ""
    args = []
    locals = bytes()


class Panels(database.Struct):
    save_by = "user_id"
    user_id = ""
    panel_name = ""


class AbstractAfterFunc:
    def __new__(cls, name, func=None, circular=None):
        if (instance := after_func_poll.get(name)) is None:
            return super().__new__(cls)
        return instance

    def __init__(self, name, func=None, circular=None) -> None:
        if getattr(self, "circular", None) is None and circular is not None:
            self.circular = circular
        if after_func_poll.get(name) is None:
            self.name = name
            self.text_matchers = {}
            after_func_poll[name] = self
        if func is not None:
            self.text_matchers["default"] = func


command_poll = []  # MutableList<Command>
after_func_poll = {}
waiters = ListX()  # populated by wait module


class CircularBuilder(object):
    def __init__(self, panels_return=None) -> None:
        if panels_return is None:
            panels_return = []
        self.panels_return = PanelsReturn(panels_return, self)
        self.invalid_handlers = {}

    def register_invalid_handler(self, name="default"):
        def func_wrap(func):
            self.invalid_handlers[name] = func
            return self
        return func_wrap

    def invalid_handler(self, botClass, name="default"):
        (get := self.invalid_handlers.get(name)) and get(botClass)


class PanelsReturn(list):
    def __new__(cls, other, parentClass):
        if isinstance(cls, PanelsReturn):
            return cls
        return super().__new__(cls)

    def __init__(self, other, parentClass):
        super().__init__(other)
        self.parentClass = parentClass

    def extend(self, other) -> CircularBuilder:
        if isinstance(other, str):
            super().append(other)
        else:
            super().extend(other)
        return self.parentClass


def _default_allow_enter(*args): return True


class Panel(object):

    panels = ListX()

    def __new__(cls, name, aliases=None, circular=None):
        if (panel := cls.find_panel(name)) is not None:
            return panel
        return super().__new__(cls)

    def get_all_names(self):
        return [self.name] + self.aliases

    @classmethod
    def find_panel(cls, name):
        return cls.panels.find(lambda it: name in it.get_all_names())

    def __init__(self, name, aliases=None, circular: CircularBuilder or NoneType = None) -> None:
        if aliases is None:
            aliases = []
        self.name = name
        self.aliases = aliases
        self.circular = circular or CircularBuilder()
        self.panels.append(self)
        self.command_poll = []
        self.allow_enter_func = _default_allow_enter

    def enter(self):
        def func_wrap(func):
            self.enter_func = func
        return func_wrap

    def run(self, botClass):
        Panels(user_id=botClass.user.id).panel_name = self.name

        return call_command(self.enter_func, botClass, botClass.args)

    def check(self, botClass):
        return self.allow_enter_func(botClass)

    def allow_enter(self):
        def func_wrap(func):
            self.allow_enter_func = func
        return func_wrap

    def check_and_run(self, botClass):
        if self.check(botClass):
            return self.run(botClass)

    def command(self, name, aliases=None, fixTypo=False):
        if aliases is None:
            aliases = []

        def func_wrap(func):
            self.command_poll.append(Command(name, aliases, fixTypo, func))
        return func_wrap


@dataclass
class Command(object):
    name: str
    aliases: list[str]
    fixTypo: bool
    callable: Callable


def command(name, fixTypo=True, aliases=None):
    if aliases is None:
        aliases = []

    def func_wrap(func):
        command_poll.append(Command(name, aliases, fixTypo, func))

    return func_wrap


menu = Panel("меню", aliases=["start", "бот", "старт", "начать"])


def after_func_from_lambda(name, func):
    """Generates a after function from lambda

    Args:
        name (str): name of after function
        func (function): lambda function to generate after_func from
    """
    after_func(name)(func)


def after_func(name, circular=None):
    """
    The after_func function is a decorator that takes the name of after function as an argument. 

    :param name: Used to identify after function.
    :return: A function object which is assigned to func_wrap.
    """
    def func_wrap(func):
        AbstractAfterFunc(name, func, circular)
    return func_wrap


def after_text_matcher(name, text):
    """
    The after_text_matcher function is a decorator that can be used to mark functions as 
    the handler for some text. The decorated function should take two arguments, the name of 
    the matcher and the text that matched. For example:

        @after_text_matcher('foo', 'bar')
        def handle_foo(self):
            print("I'm after function 'foo' that was executed on text match 'bar'")

        # This will cause handle_foo to be called if "bar" was the text of message and after func was set to "foo"

    :param name: Used to Identify the function.
    :param text: Used to Store the text that is to be matched.
    :return: Wrapper around function.
    """
    def func_wrap(func):
        function = AbstractAfterFunc(name)
        function.text_matchers[text] = func
    return func_wrap


def set_after(name, uID, args=None):
    """Binds after function to specific user

    Args:
        name (str): Name of after function.
        uID (str): user id.
        args (list[Any], optional): Args to be passed when after function will be executed. Defaults to None.
    """
    if args is None:
        args = []
    struct = AfterFunc(after_name=name, user_id=uID)
    struct.after_name = name
    struct.args = args


def call_command(function, *args):
    function_args = inspect.getfullargspec(function).args
    if len(function_args) == 1:
        return function(args[0])
    function(*args)


def emulate_command(botClass, command, user_id=None, after_func=None, after_args=None):
    if user_id is not None:
        from . import user
        botClass.user = user.User(user_id)
        if after_func is not None:
            set_after(after_func, user_id, after_args)
    botClass.init_text(command)
    execute_command(botClass)


def execute_command(botClass):
    selected = AfterFunc(create_new=False, user_id=botClass.user.id)
    selected_panel = Panels(create_new=False, user_id=botClass.user.id)
    _p = None
    if selected_panel is not None:
        _p = Panel.find_panel(selected_panel.panel_name)
        command_poll = _p.command_poll
    else:
        command_poll = globals()["command_poll"]

    if selected is not None and selected.after_name != "null":
        tmpAfterName = selected.after_name
        selected.after_name = "null"
        doNotReset = False
        if tmpAfterName in after_func_poll:
            after_func: AbstractAfterFunc = after_func_poll[tmpAfterName]
            if getattr(after_func, "circular", None) is not None:
                for panel in after_func.circular.panels_return:
                    _p: Panel = Panel.find_panel(panel)
                    if not _p:
                        continue
                    if botClass.text in _p.get_all_names() and _p.check(botClass):
                        _p.run(botClass)
                        return

            call = after_func.text_matchers.get(
                botClass.text) or after_func.text_matchers.get("default")
            if isinstance(call, CodeType) and selected.locals:
                waiter = waiters.find(lambda it: it.name == tmpAfterName)
                call = eval(call, pickle.loads(selected.locals), waiter.module.__dict__)
            if call is not None:
                doNotReset = call(botClass) if not selected.args else call(
                    botClass, selected.args)
                doNotReset = doNotReset[-1] if isinstance(
                    doNotReset, Iterable) else doNotReset
                if doNotReset is None or not isinstance(doNotReset, bool):
                    doNotReset = False
                if doNotReset:
                    selected.after_name = tmpAfterName
        return
    tmpCmd = []
    # loop over arguments
    if not _p:
        _p = Panel.find_panel(botClass.text)
        if _p and _p.check(botClass):
            _p.run(botClass)
            return
    prepared = None
    for i in botClass.txtSplit:
        tmpCmd.append(i)
        name = " ".join(tmpCmd)
        # prefer names over fixed commands
        for cmd in command_poll:
            if cmd.name == name or name in cmd.aliases:
                prepared = partial(call_command, cmd.callable,
                                   botClass, botClass.args)
                break
        for cmd in command_poll:
            if not cmd.fixTypo:
                continue
            matches = difflib.get_close_matches(name, cmd.aliases, cutoff=0.7)
            if not matches:
                continue
            prepared = partial(call_command, cmd.callable,
                               botClass, botClass.args)
            break

    if prepared is not None:
        prepared()
        return

    if _p is not None:
        for ret in _p.circular.panels_return:
            __p: Panel = Panel.find_panel(ret)
            if not __p:
                continue
            if botClass.text in __p.get_all_names() and __p.check(botClass):
                __p.run(botClass)
                break
        else:
            _p.circular.invalid_handler(botClass)



import json
import re

import vk_api
from requests.adapters import HTTPAdapter, Retry
from vk_api.longpoll import VkEventType, VkLongPoll
from types import LambdaType, NoneType
from .database import config
from . import cmd, database, events, imports, user, wait
from jsonxx import ListX
from .thread import Thread

imports.ImportTools(["Structs"])


class LongPoll(VkLongPoll):
    """Custom class for longpoll listening with preventing connection break errors"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.retry = Retry(connect=0, backoff_factor=0.33)
        self.http_adapter = HTTPAdapter(max_retries=self.retry)
        self.session.mount('https://', self.http_adapter)


class AbstractChatLongPoll(Thread):
    def __init__(self, config, api_class=vk_api.VkApi, **kwargs) -> None:
        self.config = config
        self.vk_session = api_class(token=config["vk_api_key"])
        self.longpoll = LongPoll(self.vk_session)
        self.vk = self.vk_session.get_api()
        self.db = database.Database(self.config)
        super().__init__(**kwargs)

    def parse_attachments(self):
        """
        The parse_attachments function parses the attachments from a message and appends them to the list of attachments.
        The function takes in a list of dictionaries, each dictionary containing an attachment type and its corresponding 
        attachment data. The function then iterates through each dictionary, extracting the attachment type and accessing 
        the corresponding data (owner_id, attachment_id, access_key). If there is no access key 
        for that particular attachment (i.e. attachment is public), then only the owner_id & attachment_id are used to create an 
        attachment string
        """
        for attachment_list in self.attachments_last_message:
            attachment_type = attachment_list['type']
            attachment = attachment_list[attachment_type]
            access_key = attachment.get("access_key")
            if attachment_type == "link":
                if attachment["url"] not in self.raw_text:
                    prepared = attachment["url"]
                    self.raw_text += f"\n{prepared}" if self.text else f"{prepared}"
                    # reset text because it sinks
                    self.init_text(self.raw_text)
                continue
            if attachment_type != "sticker":
                self.attachments.append(
                    f"{attachment_type}{attachment['owner_id']}_{attachment['id']}") if access_key is None \
                    else self.attachments.append(
                    f"{attachment_type}{attachment['owner_id']}_{attachment['id']}_{access_key}")
            else:
                self.sticker_id = attachment["sticker_id"]

    def write(self, user_id, *args, **kwargs):
        """
        The write function writes message to some user
        :param self: Used to Reference the class object.
        :param user_id: user to write
        :param *args: Used to Send a non-keyworded variable length argument list to the function.
        :param **kwargs: Used to Specify a variable number of keyword arguments.
        """
        user.User(user_id, vk=self.vk).write(*args, **kwargs)

    def reply(self, *args, **kwargs):
        """
        The reply function is used to reply to a message. It takes the following arguments:
            - *args: Args for User.write method.
            - **kwargs: A dictionary of additional attributes for User.write method.
        :param self: Used to Access variables that belongs to the class.
        :param *args: Used to Send a non-keyworded variable length argument list to the function.
        :param **kwargs: Used to Pass a dictionary of keyword arguments.
        :return: The result of the function call.
        """
        return self.user.write(*args, **kwargs)

    def on_message(self, event):
        """
        The on_message function is our bot's handler for the 'message' event, which
        is triggered whenever a message is sent. We can access the
        message that was sent using the `event` argument passed to us.

        :param self: Used to Access variables that belongs to the class.
        :param event: Message event.
        """
        pass

    def wait(self, text: LambdaType | str, lmbd : LambdaType | NoneType = None):
        wait.wait(self.user.id, lmbd) if isinstance(text, str) else wait.wait(self.user.id, text)

    def init_text(self, raw_text):
        self.raw_text: str = raw_text
        self.text = self.raw_text.lower()
        self.txtSplit = self.text.split()
        self.command = self.txtSplit[0] if len(
            self.txtSplit) > 0 else ""
        self.args = self.txtSplit[1:]

    def run(self):
        """
        The run function is the main function of the bot. It is called on thread start.
        """
        for event in self.longpoll.listen():
            if event.type == VkEventType.MESSAGE_NEW and event.to_me and not event.from_me and not event.from_group and not event.from_chat:
                self.db.begin_changes()
                self.attachments = ListX()
                self.sticker_id = None
                self.user = user.User(event.user_id, vk=self.vk)
                self.init_text(event.message.strip())
                self.event = event
                if (payload := getattr(self.event, "payload", None)) is not None:
                    self.event.payload = json.loads(payload)
                self.messages = self.user.messages.getHistory(count=3)["items"]
                self.last_message = self.messages[0]
                self.attachments_last_message = ListX(
                    self.last_message["attachments"])
                self.parse_attachments()
                try:
                    self.on_message(event)
                except Exception as e:
                    raise e
                finally:
                    self.db.end_changes()


class BotLongPoll(AbstractChatLongPoll):
    def on_start(self):
        """Emits start event"""
        events.emit("start")
        self.started = True

    def __init__(self, c=None, **kwargs) -> None:
        super().__init__(c or config, api_class=vk_api.vk_api.VkApiGroup, **kwargs)
        imports.ImportTools(["packages", "Structs"])
        self.group_id = "-" + re.findall(r'\d+', self.longpoll.server)[0]

    def set_after(self, x, y=None):
        if y is None:
            y = []
        cmd.set_after(x, self.user.id, y)

# for future uses?


class UserLongPoll(AbstractChatLongPoll):
    def __init__(self, config, **kwargs) -> None:
        super().__init__(config, api_class=vk_api.VkApi, **kwargs)



import vk_api
from vk_api import VkApi
from vk_api.exceptions import ApiError

from . import thread
from .cmd import set_after
from .keyboard import Keyboard
from .methodExecutor import MethodExecutor


class User(MethodExecutor):
    """Class representing event user"""
    # method to insert keyword map
    user_id_methods = {
        "messages.getHistory": "user_id", "users.get": "user_ids"}

    def __new__(cls, user_id, fields=None, **kwargs):
        """
        The __new__ function is the method called before __init__, it's used to create an instance of a class.
        It takes in cls (the class) as the first argument and returns an instance of that class.
        The __new__ function is also allowed to modify attributes on cls before returning the new object.

        :param cls: Used to Refer to the class, not the instance.
        :param user_id: Used to Specify the user whose data we want to get.
        :param fields=None: Used to Get the fields from the vk api.
        :param **kwargs: Used to Pass parameters to the __init__ function.
        :return: A user instance if user_id was a valid user id, otherwise None.
        """
        try:
            if fields is None:
                fields = "photo_id,photo_50"
            vk = kwargs.get("vk")
            instance = super(User, cls).__new__(cls)
            if vk is None:
                vk = thread.main_thread.vk
                kwargs["vk"] = vk
            elif isinstance(vk, VkApi):
                vk = vk.get_api()
            instance.vk = vk
            get = vk.users.get(user_ids=user_id, fields=fields)[0]
            instance.request = get
            return instance
        except:
            return None

    def __init__(self, user_id=None, fields=None, vk=None):
        """
        :param self: Used to Refer to the object itself.
        :param user_id=None: Used to Specify the user id.
        :param fields=None: Used to Specify the fields that will be returned by the api.
        :param vk=None: Used to Pass the vk object to the class.
        """
        self.avatar = f"photo{self.request['photo_id']}" if self.request.get(
            'photo_id') is not None else ""
        self.user_name = f"{self.request['first_name']} {self.request['last_name']}"
        self.id = str(self.request["id"])

        super().__init__(self.vk, self.on_method_execute)

    def write(self, message, keyboard=None, after=None, after_args=None, **kwargs):
        """Method to write a user

        Args:
            message (str): message content
            keyboard (:class:`Keyboard`, optional): Keyboard to send.
            after (str, optional): After function to ser.
            after_args (str, optional): Arguments for after function.
        """
        if (att := kwargs.pop("attachments", None)) is not None:
            kwargs["attachment"] = att
        if keyboard is not None:
            kwargs["keyboard"] = Keyboard(keyboard)
        try:
            self._vk.messages.send(user_id=self.id, message=message, random_id=vk_api.utils.get_random_id(),
                                   **kwargs)
        except ApiError:
            pass
        finally:
            if after is not None:
                set_after(after, self.id, after_args)

    def on_method_execute(self, method, kwargs):
        if method in User.user_id_methods:
            kwargs[User.user_id_methods[method]] = self.id

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, type(self)):
            raise NotImplementedError
        return self.id == o.id

    def __hash__(self) -> int:
        return int(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @property
    def mention(self):
        return f"@id{self.id}({self.user_name})"



"""Custom SQLite alchemy"""

import json
import os
import re
import sqlite3
from operator import getitem
from sqlite3 import Row
from sys import version_info
from typing import Any, AnyStr, Iterable
from warnings import warn
import jsonxx

from . import thread, timeExtension
from jsonxx.listx import ListX


def getter(x: Any, attr: AnyStr): return getattr(x, attr, x)


SUPPORTS_DROP_COLUMNS = version_info.major >= 3 and version_info.minor > 9


DEFAULT_CONFIG = """
{
    "db_file": "data/db.sqlite3",
    "db_backups": false,
    "db_backups_folder": "backups/",
    "db_backup_interval": 43200,
    "sync_timezone": "Europe/Moscow",
    "vk_api_key": ""
}
"""


def on_create_config():
    print("vk_sdk config created. Make sure to fill vk_api_key before running long pooling code. Happy coding!")
    os._exit(0)


config = jsonxx.load_advanced(
    "config.json", content=DEFAULT_CONFIG, createCallback=on_create_config)


def attrgetter(x: Any): return getter(x, "value")


def _formAndExpr(baseSql, argsList=None, getattrFrom=None, add=None):
    """
    The formAndExpr function takes a baseSql string, an argsList list, and a getattrFrom object. It then adds the attributes from the add list to the baseSql string and appends their values to argsList.
    Used internally from module

    :param baseSql: Used to Store the sql query that will be used to update the database.
    :param argsList: Used to Store the values of the parameters in add.
    :param getattrFrom: Used to Get the values from the object.
    :param add: Used to Add the column names to the basesql string.
    :return: A string that is a concatenation of the strings in add, each followed by an "and" and surrounded by open and close parentheses.
    """
    method = getitem if isinstance(getattrFrom, dict) else getattr
    if argsList is None:
        argsList = ListX()
    if any([x is None for x in (getattrFrom, add)]):
        raise ValueError
    for i, k in enumerate(add):
        baseSql += f"{k}=?"
        argsList.append(method(getattrFrom, k))
        if i != len(add) - 1:
            baseSql += " and "
    return baseSql, argsList


def to_sneak_case(string):
    """
    The to_sneak_case function converts a string to snake_case.

    Args:
        string (str): The input string. 
    Returns:
        str: The output snake_case formatted string.

    :param string: String that will be converted to sneak case.
    :return: A string in sneak case.
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


def convert_to_list_if_needed(element):
    if not isinstance(element, list):
        return ListX([element])
    else:
        return ListX(element)


class Struct:
    table_map = {}

    @classmethod
    def extract_table_name(cls):
        """Searches for table_name in Struct"""
        if hasattr(cls, "table_name"):
            return getattr(cls, "table_name")
        else:
            return to_sneak_case(cls.__name__)

    @classmethod
    def extract_save_by(cls):
        """
        The extract_save_by function is used to extract the save_by attribute from a class.

        If the class has no save_by attribute, it will look for any attributes that are instances of Sqlite3Properties that has unique modifier in their "value".

        :param cls: Used to Access the class object of the current instance.
        :return: A list of the names of all the attributes that are marked for saving.
        """
        if hasattr(cls, "save_by"):
            return convert_to_list_if_needed(attrgetter(cls.save_by))
        for k, v in vars(cls).items():
            if isinstance(v, Sqlite3Property) and "unique" in v.type:
                return ListX(k)

    def __getstate__(self):
        return _formAndExpr(f"select * from {self.table_name} where ", [], self, self.save_by)

    def __setstate__(self, t):
        return db.select_one_struct(t[0], t[1], fromSerialized=self)

    def __init_subclass__(cls) -> None:
        """
        The __init_subclass__ function is called when some class derives from our Struct
        It`s purpose is to ensure that all subclasses of the base class have their own table_name and save_by attributes, 
        and that they are kept in sync with the base class via an updated table_map.

        :param cls: Used to Access the class object.
        :return: The class object that is being defined.
        """
        cls.table_name = cls.extract_table_name()
        cls.save_by = cls.extract_save_by()
        cls.table_map[cls.table_name] = cls
        return super().__init_subclass__()

    def __new__(cls, **kwargs):
        create_new = kwargs.get("create_new", True)
        instance = super().__new__(cls)
        instance.setattr("old_struct", None, ignore_duplicates=False)
        if kwargs:
            expr = f"select * from {cls.table_name} where "
            expr, args = _formAndExpr(
                expr, getattrFrom=kwargs, add=cls.save_by)
            old_struct = cls.db.select_one_struct(expr, args)
            if old_struct is None and create_new is False:
                return None
            instance.old_struct = old_struct
        return instance

    def __init__(self, create_new=None, **kwargs) -> None:
        if kwargs:
            for key, value in kwargs.items():
                setattr(self, key, value)
            if all(self.save_by.map(lambda it: it in kwargs.keys())):
                if self.old_struct is None:
                    keys, values = ListX(), ListX()
                    vrs = self.vars()
                    for key, value in vrs.items():
                        keys.append(key)
                        values.append(kwargs[key]) if kwargs.get(
                            key) is not None else values.append(value)
                    insert_string = f"insert or ignore into {self.table_name} ({','.join(keys)}) values ({values.map(lambda _: '?').join(',')})"
                    self.db.execute(insert_string, values)
                    variables = self.vars()
                    self.fill(variables.keys(), variables)
                else:
                    keys = kwargs.keys()
                    values = ListX(kwargs.values()).find_all(
                        lambda it: it not in self.save_by)
                    d = dict(zip(keys, values))
                    for k, v in d.items():
                        self.old_struct.setattr(k, v)
                    variables = self.old_struct.vars()
                    self.fill(variables.keys(), variables)
                self.initialized = True

        super().__init__()

    @classmethod
    def select_all(cls, **kwargs):
        if len(kwargs) > 0:
            expr, args = _formAndExpr(
                f"select * from {cls.table_name} where ", getattrFrom=kwargs, add=kwargs)
            return cls.db.select_all_structs(expr, args)
        else:
            return cls.db.select_all_structs(f"select * from {cls.table_name}")

    @classmethod
    def row_at(cls, num):
        return cls.db.select_one_struct(f"select * from {cls.table_name} limit 1 offset {num}")

    @classmethod
    def first(cls): return cls.row_at(0)

    def boundStructByAction(self, key, data):
        """Bounds struct by action to a given data (list or dict). Struct by action will handle the watching on elements change."""
        structByAction = jsonxx.JsonX(data, saver=lambda _: self.db.save_struct_by_action(_,
                                                                                                          self.table_name, key, self.save_by, self))
        return structByAction

    def destroy(self):
        """
        The destroy function deletes Struct record from db.
        """
        sql = f"delete from {self.table_name} where "
        sql, lst = _formAndExpr(sql, getattrFrom=self, add=self.save_by)
        self.db.execute(sql, lst)

    def setattr(self, key, value, write_to_database=True, ignore_duplicates=True):
        """
        The setattr function is a helper function that allows us to write the value of an attribute to the database.
        It is called when we set an attribute on a Struct, and it will only write to the database if:
        - The struct has been initialized (i.e., it's ready to use and that's not some internal module call)
        - The key being assigned matches one of our attributes (otherwise, we might be trying to assign something that doesn't correspond with any known field in our struct)
        - The value being assigned is different from what was previously stored for this key in our struct (unless ignore_duplicates=True). This prevents us from writing unnecessary updates.
        Otherwise it's some internal package calls.

        :param key: Used to Specify the attribute that is to be set.
        :param value: Used to Set the value of an attribute.
        :param write_to_database=True: Do we need to write changed value to db.
        :param ignore_duplicates=True: Used to Avoid writing the same value to the database multiple times.
        :return: None.
        """
        prev = getattr(self, key, None)
        if prev == value and ignore_duplicates:
            return
        if write_to_database and getattr(self, "initialized", False):
            if isinstance(prev, tuple(jsonxx.ExtensionBase.classes.values())):
                # resave
                super().__setattr__(key, self.boundStructByAction(key, value))
                getattr(self, key).save()
            else:
                self.db.write_struct(self, key, value)
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def vars(cls):
        """
        The vars function is used to extract the class variables from a class.

        :param cls: Used to Indicate the class that we want to get attributes from.
        :return: A dictionary of the class's namespace.
        """
        attrs = {k: getattr(cls, k) for k in dir(cls)}
        return {k: v for k, v in attrs.items() if not k.startswith(
                "__") and k not in ["table_name", "save_by", "initialized", "table_map", "use_db", "db", "old_struct"] and (not callable(v) or type(v) in jsonxx.ExtensionBase.classes)}

    def fill(self, keys, getitemfrom):
        """Fills attributes mapped from list[str] keys to getitemfrom object to our Struct"""
        for k in keys:
            value = jsonxx.JsonX.accept(getitemfrom[k], saver=lambda _: self.db.save_struct_by_action(_,
                                                                                                                     self.table_name, k, self.save_by, self))

            if isinstance(getattr(self, k), bool):
                value = self.str2bool(value)

            self.setattr(k, value, False)

    def __setattr__(self, name: str, value: Any) -> None:
        self.setattr(name, value)

    def str2bool(self, v):
        if hasattr(v, "lower"):
            return v.lower() in ("true", "1")
        return v

    def __repr__(self) -> str:
        return f"{self.table_name}({', '.join([f'{x}={y}' for x, y in self.vars().items()])})"


class Sqlite3Property(object):
    def __init__(self, x: Any, y: AnyStr):
        self.value = x
        self.type = y


class Database:

    typeToTypeCache = {str: "text", dict: "str", list: "str",
                       float: "real", type(None): "null", int: "int", bool: "bool", bytes: "blob"}
    db_cache = {}

    def __new__(cls, settings, **kwargs):
        file = settings["db_file"]
        if (instance := cls.db_cache.get(file)) is None:
            instance = super().__new__(cls)
            db_short = os.path.basename(
                os.path.splitext(file)[0])
            cls.db_cache[db_short] = instance  # short path
            cls.db_name = db_short  # instance.db_name ?
            cls.db_cache[file] = instance
            return instance
        return instance

    def end_changes(self):
        if self.need_commit and self.stash:
            self.db.commit()
            self.need_commit = False
            self.stash = False

    def begin_changes(self):
        self.stash = True

    def __init__(self, settings: dict, **kwargs):
        self.settings = settings
        self.backup_folder = self.settings["db_backups_folder"]
        folder = os.path.split(self.settings["db_file"])[0]

        if not os.path.exists(folder):
            os.makedirs(folder)
        self.file = self.settings["db_file"]
        self.db = sqlite3.connect(
            self.settings["db_file"], check_same_thread=False, **kwargs)
        self.row_factory = sqlite3.Row
        self.db.row_factory = self.row_factory
        self.cursor = self.db.cursor()
        self.stash = False
        self.need_commit = False

        is_main_db = self.settings["db_file"] == config["db_file"]

        if is_main_db:
            global db
            db = self

        for struct in Struct.table_map.values():
            if (not is_main_db and not hasattr(struct, "use_db")) or (hasattr(struct, "use_db") and self.db_cache[struct.use_db] != self):
                continue
            iterable = -1
            rows = []
            struct.db = self
            variables = struct.vars(struct)
            for key, value in variables.items():
                iterable += 1
                real_value = attrgetter(value)
                rows.append(
                    f"{key} {self.convert_type(real_value)} {getattr(value, 'type', '')} default \"{real_value}\"")
            self.execute(
                f"create table if not exists {struct.table_name} ({', '.join(rows)})")
            table_fields = self.get_column_names(struct.table_name)
            for iterable, field in enumerate(variables.keys()):
                if field not in table_fields:
                    self.execute(
                        f"alter table {struct.table_name} add column {rows[iterable]}")
            skipped_drop = ListX()
            for field in list(table_fields):
                if field not in variables:
                    if SUPPORTS_DROP_COLUMNS or config.get("disable_column_drop_checks"):
                        self.execute(
                            f"alter table {struct.table_name} drop column {field}")
                    else:
                        skipped_drop.append(
                            f"{field} (struct {struct.table_name})")
            if len(skipped_drop) > 0:
                warn(f"There are {len(skipped_drop)} columns waiting to be deleted from database {self.db_name}." +
                     f"Unfortunately, your python version doesn't support this sqlite3 operation. Please, delete the following fields by yourself:\n" +
                     f"{''.join(skipped_drop)}\n\n(You can try setting disable_column_drop_checks to True in config file and see if it works)")
        if self.settings["db_backups"]:
            thread.every(self.settings["db_backup_interval"], name="Backup")(
                self.backup)

    def backup(self):
        if not self.settings["db_backups"]:
            return

        rawName = self.file.split("/")[-1]
        manager = thread.ThreadManager()
        manager.changeInterval(
            "Backup", self.settings["db_backup_interval"])
        backup_table = sqlite3.connect(
            f"{self.backup_folder}backup_{timeExtension.Timestamp.now()}_{rawName}")
        self.db.backup(backup_table)

    def select(self, query: AnyStr, args=None):
        if args is None:
            args = []
        self.cursor.execute(query, args)
        return self.cursor.fetchall()

    def select_one(self, query: AnyStr, *args):
        if isinstance(args, list):
            self.cursor.execute(query, [str(x) for x in args])
        else:
            self.cursor.execute(query, *args)
        return self.cursor.fetchone()

    def write_struct(self, structToWrite: Struct, changedKey: AnyStr, newValue: Any):
        table = structToWrite.table_name
        unique_fields = Struct.table_map[table].save_by
        sql = f"update or ignore {table} set {changedKey} = ? where "
        argsList = [newValue]
        sql, argsList = _formAndExpr(
            sql, argsList, structToWrite, unique_fields)
        self.execute(sql, argsList)

    def select_one_struct(self, query: AnyStr, *args: tuple or jsonxx.JsonX,
                          selectedStruct: Row = None,
                          fromSerialized=None, table_name=None):
        table_name = self.parse_table_name(query, table_name)
        struct = self.select_one(
            query, *args) if selectedStruct is None else selectedStruct
        if struct is None:
            return None
        if not isinstance(table_name, str):
            raise Exception(
                f"Table name's type is not string (table_name was not provided correctly?)\n{query=}\n{args=}\n{table_name=}")
        myStruct: Struct = Struct.table_map[table_name](
        ) if fromSerialized is None else fromSerialized
        if struct is None:
            return None
        myStruct.fill(struct.keys(), struct)
        myStruct.setattr("initialized", True, write_to_database=False)
        return myStruct

    def select_all_structs(self, query: AnyStr, *args) -> ListX:
        structs = ListX(self.select(query, *args))
        return ListX([self.select_one_struct(query, *args, selectedStruct=x) for x in structs])

    def save_struct_by_action(self, value, table_name: AnyStr, key: Any,
                              unique_field: Iterable, parent_struct: Struct):
        baseSql = f"update {table_name} set {key} = ? where "
        argsList = [value]
        baseSql, argsList = _formAndExpr(
            baseSql, argsList, parent_struct, unique_field)
        self.execute(baseSql, argsList)

    def execute(self, query: AnyStr, args=None):
        if args is None:
            args = []
        for i, k in enumerate(args):
            if isinstance(k, tuple(jsonxx.ExtensionBase.classes.values())):
                args[i] = jsonxx.JsonX(k).dumps()
        self.cursor.execute(query, args)
        if self.stash:
            self.need_commit = True
        else:
            self.db.commit()
        return self.cursor

    def get_column_names(self, table: AnyStr):
        select = self.cursor.execute(f"select * from {table}")
        return [x[0] for x in select.description]

    def parse_table_name(self, query, fromCached=None):
        if fromCached is None:
            return list(tables_in_query(query))[0]
        return fromCached

    def get_table_names(self):
        return [x["name"] for x in self.select("SELECT name FROM sqlite_master WHERE type='table'")]

    @staticmethod
    def convert_type(value):
        return Database.typeToTypeCache[type(value)]


db: Database = None


# https://grisha.org/blog/2016/11/14/table-names-from-sql/
def tables_in_query(sql_str):
    q = re.sub(r"/\*[^*]*\*+(?:[^*/][^*]*\*+)*/", "", sql_str)
    lines = [line for line in q.splitlines(
    ) if not re.match(r"^\s*(--|#)", line)]
    q = " ".join([re.split(r"--|#", line)[0] for line in lines])
    tokens = re.split(r"[\s)(;]+", q)
    result = set()
    get_next = False
    for tok in tokens:
        if get_next:
            if tok.lower() not in ["", "select"]:
                result.add(tok)
            get_next = False
        get_next = tok.lower() in ["from", "join"]

    return result



from vk_api import VkApi
from vk_api.exceptions import ApiError


def empty(*args, **kwargs):
    """
    The empty function

    :param *args: Used to Pass a non-keyworded, variable-length argument list.
    :param **kwargs: Used to Pass a keyworded, variable-length argument list.
    :return: "none".
    """
    pass


class MethodExecutor(object):
    def __init__(self, vk, on_method_execute, on_error=None, method=None) -> None:
        if isinstance(vk, VkApi):
            vk = vk.get_api()
        self._method = method
        self.on_method_execute = on_method_execute or empty
        self.on_error = on_error or empty
        self._vk = vk

    def __getattr__(self, method):
        """
        The __getattr__ function is called when an attribute is not found in the usual places 
        (__dict__, class tree). It can bebe used to delegate the lookup to another object (usually either 
        the superclass or a mixin) and/or try some dynamic approach (e.g., see this recipe).

        :param self: Used to Access the class instance from within the method.
        :param method: Used to Determine which method is called.
        :return: A methodexecutor object.
        """
        if '_' in method:
            m = method.split('_')
            method = m[0] + ''.join(i.title() for i in m[1:])
        return MethodExecutor(
            self._vk,
            self.on_method_execute,
            self.on_error,
            (self._method + '.' if self._method else '') + method
        )

    def __call__(self, **kwargs):
        if self._method is not None:
            if self.on_method_execute is not None:
                shouldExecute = self.on_method_execute(self._method, kwargs)
                shouldExecute = shouldExecute or True
            if not shouldExecute:
                return
            self._vk._method = self._method
            try:
                tmpReturn = self._vk.__call__(**kwargs)
                return tmpReturn
            except Exception as e:
                self.on_error(e)
            finally:
                self._vk._method = None
                self._method = None


class AuthBasedMethodExecutor(MethodExecutor):
    def __init__(self, token, invalid_callback=None) -> None:
        """
        The __init__ function is called when a new instance of the class is created. 
        It initializes all of the variables in the class and can accept arguments that 
        are passed to it at creation time. In this case, we are creating an object that 
        contains a token and an invalid_callback function.

        :param self: Used to Reference the class instance itself.
        :param token: Used to Authenticate the user through VK.
        :param invalid_callback=None: Used to Specify a function to call when the token is invalid.
        :return: None.
        """
        self._valid = True
        self.invalid_callback = invalid_callback or empty
        vk = VkApi(token=token).get_api()
        super().__init__(vk, self.on_method_execute, self.on_error)

    def on_method_execute(self, *args, **kwargs):
        return self._valid

    def on_error(self, e):
        if isinstance(e, ApiError) and e.code == 5:
            self._valid = False
            self.invalid_callback()



import random

from vk_api import keyboard


class Keyboard(keyboard.VkKeyboard):
    colors = {"blue": "primary", "white": "secondary",
              "red": "negative", "green": "positive"}

    def __new__(cls, buttons=None, **kwargs):
        if isinstance(buttons, cls):
            return buttons
        return super().__new__(cls)

    def __init__(self, buttons=None, strategy="default", **kwargs):
        """
        The __init__ function is called when a new object is created from the class.
        The __init__ function can take arguments, but self is always the first one.
        Self is just a reference to the instance of the class. It's automatically 
        passed in when you instantiate an instance of the class.

        :param self: Used to Refer to the object itself.
        :param buttons=None: Dict/List/str describing buttons.
            could be Dict:
                        name       color
                :code:`{"button1": "red"}`
            could be List:
                         name
                :code:`["button1", "button2"]`
            could be str:
                         name
                :code:`"button_name"`
            for list and str we use random colors
        :param strategy="default": Used to Set the default strategy for handling append on the buttons.
        :param one_time=False: Used to Specify if the keyboard should be removed after one response.
        :param inline=False: Used to Specify whether the buttons should be bounded to a message.
        :return: The object of the class.
        """
        if hasattr(self, "lines") and len(self.lines) > 0:
            return
        if buttons is None:
            buttons = []
        self.strategy = strategy
        super().__init__(**kwargs)
        self.button_index = 0
        self.__add__(buttons)

    def get_keyboard(self):
        """Cleanups empty lines and returns keyboard as json"""
        self.cleanup()
        return super().get_keyboard()

    __repr__ = get_keyboard

    @classmethod
    def empty(cls):
        return cls()

    def add_button(self, *args, **kwargs):
        if getattr(Strategies, self.strategy)(self):
            self.add_line()
        args = list(args)
        if len(args) >= 2:
            color = args[1]
            del args[1]
        else:
            color = kwargs.get("color")
        if color is not None:
            if color in self.colors:
                color = self.colors[color]
        else:
            color = self.colors["white"]
        kwargs["color"] = color
        super().add_button(*args, **kwargs)
        self.button_index += 1
        return self

    def cleanup(self):
        for i, line in enumerate(self.lines):
            if not line:
                del self.lines[i]

    def add_line(self):
        """
        The add_line function adds a line to the Keyboard.
        It also removes any empty lines from the Keyboard.

        :param self: Used to Refer to the object itself.
        :return: None.
        """
        self.cleanup()
        self.button_index = 0
        super().add_line()
        return self

    def add_from_list(self, arr):
        for j in arr:
            if j != "locationButton":
                self.add_button(j, color=self.random_color())
            else:
                self.add_location_button()
                self.add_line()

    def add_from_dict(self, d):
        if (strategy := d.get('strategy')) is not None:
            self.strategy = strategy
            del d['strategy']
        for k, v in d.items():
            if v == "line":
                self.add_line()
            else:
                self.add_button(k, v)

    def random_color(self):
        """
        The get_random_color function returns a random color.

        :returns: A random color from `Keyboard.colors`:

        :param self: Used to Access variables that belongs to the class.
        """
        return random.choice(list(self.colors))

    def __add__(self, other):
        if isinstance(other, dict):
            self.add_from_dict(other)
        elif isinstance(other, list):
            self.add_from_list(other)
        elif isinstance(other, str):
            self.add_button(other, color=self.random_color())
        return self

    __iadd__ = __add__


class Strategies:
    @staticmethod
    def default(keyboard: Keyboard):
        return keyboard.button_index % 4 == 0

    @staticmethod
    def insert_lines(_):
        return True

    @staticmethod
    def max_two(keyboard: Keyboard):
        return keyboard.button_index % 2 == 0

    max_two_buttons = max_two

    @staticmethod
    def max_five(keyboard: Keyboard):
        return keyboard.button_index % 5 == 0




class Event(object):
    """Custom class for event handling"""
    callbacks = {}

    def __init__(self, name, func, param="") -> None:
        self.name = name
        self.param = param
        self.func = func
        callbacks = self.callbacks.get(name)
        if callbacks is None:
            Event.callbacks[name] = [self]
        else:
            Event.callbacks[name].append(self)
        super().__init__()


def on(name):
    """
    The on function is a decorator that registers an event handler.
    It takes the name of the event as its first argument, and a function as its second argument.
    The decorated function will be called whenever the named event is triggered.

    :param name: Used to Identify the event.
    :return: A function that takes a function as an argument.
    """
    def func_wrap(func):
        Event(name, func, "on")
    return func_wrap


def once(name):
    """
    The once function is a decorator that will only call the decorated function once. 
    The decorated function will be called whenever the named event is triggered and then callback will be removed.

    :param name: Used to Identify the event.
    :return: A function that takes a function as an argument.
    """
    def func_wrap(func):
        Event(name, func, "once")
    return func_wrap


def emit(name, *args, **kwargs):
    callbacks = Event.callbacks.get(name)
    if callbacks is not None:
        for iterable, callback in enumerate(callbacks):
            callback.func(*args, **kwargs)
            if callback.param == "once":
                del Event.callbacks[name][iterable]



import threading

from jsonxx import listx

from . import events


class ThreadManager(object):
    thread_poll = listx.ListX()

    @classmethod
    def get_main_thread(cls):
        return cls.threadByName("Main")

    @classmethod
    def threadByName(cls, name):
        return cls.thread_poll.find(lambda item: item.name == name)

    def changeInterval(self, name, newInterval):
        thread = self.threadByName(name)
        thread.interval = newInterval

    @classmethod
    def create_task(cls, name, task, *args, **kwargs):
        thread = cls.threadByName(name)
        thread.create_task(task, *args, **kwargs)

    def __getitem__(self, key):
        return self.threadByName(key)


class Thread(threading.Thread):
    def __init__(self, *args, **kwargs) -> None:
        ThreadManager.thread_poll.append(self)
        self.tasks = []
        super().__init__(*args, **kwargs)

    def create_task(self, task, *args, **kwargs):
        self.tasks.append((task, args, kwargs))

    def check_tasks(self):
        while len(self.tasks) > 0:
            task = self.tasks.pop(0)
            task[0](*task[1], **task[2])


def requires_start(func):
    def call_wrap(*args, **kwargs):
        if main_thread is None or not getattr(main_thread, "started", False):
            @events.once("start")
            def _():
                func(*args, **kwargs)
        else:
            func(*args, **kwargs)

    return call_wrap


def threaded(*args, **kwargs):
    def func_wrap(func):
        def call_wrap(*a, **b):
            Thread(*args, **kwargs, target=func, args=a, kwargs=b).start()
        return call_wrap
    return func_wrap


class Every(Thread):

    __slots__ = ('interval', 'stopped', 'event',
                 'onExecCallback', 'args', 'callback')

    def __init__(self, interval, *args, onExecCallback=None, callback=None, **kwargs):
        if callback is not None:
            self.callback = callback
        elif hasattr(self, "loop"):
            self.callback = self.loop
        else:
            raise Exception("Callback wasn't provided.")
        self.interval = interval
        self.stopped = False
        self.event = threading.Event()
        self.onExecCallback = onExecCallback
        self.args = args
        super().__init__(**kwargs)
        self.start()

    # override
    def run(self):
        """
        The run function runs callback function. It checks for new tasks, and executes them.

        :param self: Used to Access the class instance from within the function.
        """
        self.callback(*self.args)
        while not self.event.wait(self.interval) and not self.stopped:
            if self.onExecCallback is not None:
                self.onExecCallback()
            self.check_tasks()
            self.callback(*self.args)


def every(interval, *myArgs, callback=None, **myKwargs):
    """
    The every function is a decorator that schedules the given function to be called periodically. 
    The decorated function is called with the given arguments and keyword arguments, as well as any additional 
    arguments passed when calling the Every instance. The decorated function is called at first invocation 
    and then repeatedly every interval seconds, or with an optional initial delay.

    :param interval: Used to Set the time interval between each execution of the function.
    :param *myArgs: Used to Pass arguments to the function that is being called.
    :param callback=None: Used to Pass a function to be called when the timer is executed.
    :param **myKwargs: Used to Pass in keyword arguments to the callback function.
    :return: A function that is a wrapper around the original function.
    """
    def func_wrap(func):
        return Every(interval, *myArgs, onExecCallback=callback, **myKwargs, callback=func)

    return func_wrap


main_thread = None


@events.once("start")
def update_main_thread():
    global main_thread
    main_thread = ThreadManager.threadByName("Main")



import ast
import inspect
from dataclasses import dataclass
from itertools import zip_longest
from typing import Union
from types import CodeType
import pickle

from .imports import ImportTools
from .cmd import AfterFunc, waiters, after_func_poll, AbstractAfterFunc

WAIT_FUNC_NAME = "wait"

def parent_chain(chain_in):
    chain = chain_in.split(".")
    return ".".join(chain[:-1]+[chain[-1].split("_")[0]]) if "_" in chain[-1] else chain_in

def compare_ast(node1: Union[ast.expr, list[ast.expr]], node2: Union[ast.expr, list[ast.expr]]) -> bool:
    if type(node1) is not type(node2):
        return False

    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in {"lineno", "end_lineno", "col_offset", "end_col_offset", "ctx"}:
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True

    elif isinstance(node1, list) and isinstance(node2, list):
        return all(compare_ast(n1, n2) for n1, n2 in zip_longest(node1, node2))
    else:
        return node1 == node2


@dataclass
class Waiter:
    name: str
    func: CodeType
    node: ast.Lambda
    module: object


def find_data(func):
    func = ast.parse(inspect.getsource(func).strip(),
                     mode="exec").body[0].value
    node = func.args[1] if isinstance(
        func.args[0], ast.Constant) else func.args[0]
    for waiter in waiters:
        if compare_ast(waiter.node, node):
            # retrieve locals from stack
            stack = inspect.stack()
            for index, item in enumerate(stack[2:-1]):
                inside = stack[1+index]
                if inside[3] == "wait" and not inside.filename.endswith("vk_sdk\\wait.py"):
                    locs = item[0].f_locals
                    locs.pop("self")
                    waiter = waiters.find(lambda it: it.name == parent_chain(waiter.name))
                    return locs, waiter


def wait(user_id, func):
    locs, waiter = find_data(func)

    return AfterFunc(after_name=waiter.name, user_id=user_id,
                     locals=pickle.dumps(locs))


class WaitVisiter(ast.NodeVisitor):
    def __init__(self, code, module) -> None:
        super().__init__()
        self.code = code
        self.module = module
        self.visit_Module(code)

    def dump(self, node):
        print(ast.dump(node, indent=4))

    def visit_call(self, node: ast.Call, chain: str, counter: int):
        if isinstance(node.func, ast.Attribute) and node.func.attr == WAIT_FUNC_NAME:
            assert 0 <= len(node.args) <= 3
            match = len(node.args) == 2
            if match: match_text = node.args[0].value
            lambda_func = node.args[0] if not match else node.args[1]
            current_chain = chain+f".wait{counter}{'' if not match else '_'+match_text}"

            evaluted = compile(ast.Expression(
                body=lambda_func), filename="", mode="eval")

            waiters.append(Waiter(name=current_chain,
                           func=evaluted, node=lambda_func, module=self.module))
            # circular support?
            
            pc = parent_chain(current_chain)
            af = AbstractAfterFunc(
                pc, None, None)
            if match:
                assert isinstance(node.args[0], ast.Constant)
                af.text_matchers[match_text] = evaluted
            else:
                af.text_matchers["default"] = evaluted
            after_func_poll[pc] = af
            # visit inners
            cnt = 0
            if isinstance(lambda_func.body, ast.Call):
                self.visit_call(lambda_func.body, current_chain, 0)
            elif isinstance(lambda_func.body, ast.Tuple):
                for inner in lambda_func.body.elts:
                    if isinstance(inner, ast.Call):
                        self.visit_call(inner,
                                                   current_chain, cnt)
            counter += 1 and not match

    def visit_Module(self, node: ast.Module):
        for item in node.body:
            if isinstance(item, ast.FunctionDef) and hasattr(item, "name"):
                counter = 0
                for inside in item.body:
                    if isinstance(inside, ast.Expr) and isinstance(inside.value, ast.Call):
                        self.visit_call(inside.value,
                                                   str(item.name), counter)
        self.generic_visit(node)


@ImportTools.on_module()
def on_module_ast(module, source):
    WaitVisiter(ast.parse(source), module)



"""Python importing simplified!"""

import importlib
import inspect
import os
import pathlib
from importlib import util
from types import ModuleType


class ImportOrder(object):
    orders = {}
    file_name = "__import_order__.py"

    def __init__(self, *names, sort_files=False, file=None) -> None:
        file = file or inspect.stack()[1].filename
        self.order = []
        names = [
            name + ".py" if not name.endswith(".py") else name for name in names]
        self.order += names
        names = set(names)
        parent = pathlib.Path(file).parent.resolve()
        root = pathlib.Path(os.path.abspath(os.curdir))
        rel_path = str(parent.relative_to(root))
        files = os.listdir(parent)
        if sort_files:
            files.sort()
        for file in files:
            if file not in names:
                self.order.append(file)
        self.orders[rel_path] = self

    @classmethod
    def resolve(cls, path):
        lsdir = os.listdir(path)
        if cls.file_name in lsdir:
            require(f"{path}/{cls.file_name}")
            return cls.orders[path].order
        else:
            return lsdir


class ImportTools:
    ignore = ["__pycache__"]
    imported = set()
    import_queue = set()
    modules = {}
    _on_module = set()

    @classmethod
    def on_module(cls):
        def func_wrap(func):
            cls._on_module.add(func)
        return func_wrap

    def __init__(self, paths=None):
        if paths is None:
            paths = ["packages"]
        for path in paths:
            if path in self.imported or path in self.import_queue:
                continue
            self.import_queue.add(path)
            if not os.path.exists(path):
                os.makedirs(path)
            for file in ImportOrder.resolve(path):
                if path in self.ignore:
                    continue
                thisPath = os.path.join(path, file)
                if os.path.isdir(thisPath):
                    continue
                self.imp_by_path(thisPath)
            self.import_queue.remove(path)
            self.imported.add(path)

    def reload(self, module: str):
        """
        The reload function reloads a module. It takes in the name of the module as an argument and reloads it.

        :param module: Used to Specify the module that is to be reloaded.
        :return: The module that was reloaded.
        """
        (module := self.modules.get(module)) and importlib.reload(module)

    def reload_all(self):
        """
        The reload_all function reloads all of the modules.

        :param self: Used to Access the attributes and methods of the class in which it is used.
        """
        for k, v in self.modules.items():
            self.modules[k] = importlib.reload(v)

    @classmethod
    def imp_by_path(cls, path) -> ModuleType:
        """
        The imp_by_path function imports a Python module from a path. Like a node.js' require 

        :param cls: Used to Access the class attributes.
        :param path: Used to Specify the path of the module.
        :return: The module that was loaded from the path.
        """
        if not path.endswith(".py"):
            path += ".py"
        module_name = os.path.splitext(os.path.basename(path))[0]
        spec = util.spec_from_file_location(module_name, path)
        foo = util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        cls.modules[path.replace(os.path.sep, "/")] = foo
        source = None
        for func in cls._on_module:
            if len(inspect.signature(func).parameters) == 2:
                if source is None:
                    with open(spec.origin, encoding="utf-8_sig") as f:
                        source = f.read()
                func(foo, source)
            else: func(foo)
        return foo


require = ImportTools.imp_by_path



"""Various extentions for time managing time"""

import time
from datetime import datetime

import pytz


class Timestamp(object):

    __slots__ = ('timestamp', 'diff', 'sync_tz')

    def __init__(self, timestamp=None, sync_tz="Europe/Moscow"):
        """
        :param timestamp=None: Used to Set the timestamp to the current time if it is not specified.
        :param sync_tz="Europe/Moscow": Used to Set the timezone to sync timestamp with.
        """
        self.sync_tz = sync_tz
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = timestamp
        zrh = pytz.timezone(self.sync_tz)
        tztime = zrh.localize(datetime.fromtimestamp(self.timestamp))
        tzfloat = (tztime - datetime(1970, 1, 1,
                   tzinfo=pytz.utc)).total_seconds()
        self.diff = self.timestamp - tzfloat

    def get_time(self):
        return self.timestamp + self.diff

    def __float__(self):
        return self.get_time()

    def passed(self):
        return time.time() >= self.get_time()

    @classmethod
    def now(cls):
        """Create Timestamp class from current time"""
        return cls(time.time())

    def prettyprint(self):
        """
        The prettyprint function prints the contents of a object in a human-readable format.
           It is intended to be used for debugging purposes.

        :param self: Used to Reference the object to which the function is attached.
        :return: A string representation of the object.
        """
        return datetime.fromtimestamp(self.get_time()).strftime("%Y/%m/%d %H:%M:%S")
